library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- Version 2
-- for screamer 2 SBC with Z180 and 256 colours

entity gfx is
port (
clk		:	in std_logic ;	-- 25mhz master clock VGA
reset		:	in	std_logic ;	-- system master reset
wr			:	in	std_logic ;	-- CPU Write
VBI		:	out std_logic ;	-- interupt

-- memory interface
Vaddr		:	out std_logic_vector(11 downto 0);	-- vram address lines (4k DP ram)
Vdata		:	in std_logic_vector(7 downto 0);		-- video data
-- vga output
hsync		:	out std_logic;		-- VGA Syncs
vsync		:	out std_logic;
red		:	out std_logic;		-- 8 colour bit output
red_i		:	out std_logic;
red_x		:	out std_logic;    -- no extra bit for green
green		:	out std_logic;
green_i	:	out std_logic;
blue		:	out std_logic;
blue_i	:	out std_logic;
blue_x	:	out std_logic;
-- control input
option	:	in std_logic;	-- active low write to control regiser
Raddr		:	in std_logic_vector(1 downto 0);	-- up to 4 registers
Rdata		:	in std_logic_vector(7 downto 0)	-- 8 bit colour data from CPU
) ;
end entity ;
architecture behavioral of gfx is
-- 64 characters, 8 character border, 2 character porch 16 characters sync, 
--6 character porch, 8 character border, 100 characters = 800 pixels total
constant R_border:	integer := 512;
constant H_Porch:    integer := 576;
constant H_Sync:     integer := 592;
constant H_SyncEnd:  integer := 688;
constant L_border:	integer := 736;
constant H_read:		integer := 792;
constant H_Total:    integer := 800;

signal H_Count:      integer range 0 to H_Total;
signal H_Blank:      std_logic := '0';
signal H_Border:     std_logic := '0';

-- 48 rows of 10 scan lines = 480 lines, (8 form char deff then then 2 border colour, 12 porch, 2 vertical 
-- sync then 33 porch 
constant V_Porch:    integer := 480;
constant V_Sync:     integer := 492;
constant V_SyncEnd:  integer := 494;
constant V_Total:    integer := 525;

signal V_Count:      integer range 0 to V_Total;
signal V_Blank:      std_logic := '0';

constant CharHeight: integer := 10-1;
constant RamSize:    integer := 4096-1;

signal char_data:		std_logic_vector (7 downto 0);
signal pixel_data:	std_logic_vector (7 downto 0);
signal ink_data:		std_logic_vector (7 downto 0);
signal paper_data:	std_logic_vector (7 downto 0);
signal border_data:	std_logic_vector (7 downto 0);
signal C_Count:      integer range 0 to 64-1;
signal ScanLine:     integer range 0 to CharHeight;
signal rowCount:     integer range 0 to 47;
Signal vScrollReg:    std_logic_vector (5 downto 0);
Signal hScrollReg:    std_logic_vector (5 downto 0);

signal regUpdate:		std_logic :='1' ;
signal VBIenable:		std_logic :='0' ;
signal blank_area:	std_logic :='0' ;
signal border_area:	std_logic :='0' ;
signal sequence:		integer range 0 to 7 ;	-- 8 step memory timing

begin
-- define the screen area's, sync is wholy within blank. blank is wholy within border
-- anything not border is active.
blank_area   <= '0' when v_blank ='0' or h_blank ='0' else '1';
border_area  <= '0' when ScanLine>7 or H_border ='0' or v_blank ='0' else '1';
regUpdate	<= '0' when option='0' and wr='0' else '1';

-- VBI active (low) during line 480 of the display IF VBI enable is set by a writing a 1
-- to register 3 bit 6
VBI <='0' when (v_count = V_porch) AND VBIenable ='1' else '1' ;

MemoryControl:
Process(clk,H_count)
begin
	IF rising_edge(clk) then
		if (h_count = h_read) then       -- first memory access starts 8 clocks before the active area
			c_count <= to_integer(unsigned(hScrollReg)) ;       
		end if ;
		sequence <= h_count mod 8;
		case sequence is
			when 0 =>		-- put the character cell address onto the address bus
				Vaddr(11 downto 6) <= std_logic_vector(to_unsigned(rowcount,6));
				Vaddr(5 downto 0) <= std_logic_vector(to_unsigned(c_count,6));
				pixel_data(7 downto 1) <= pixel_data(6 downto 0) ; 
			when 1 =>
				pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
			when 2 =>		-- give the ram time to respond before reading in the data
				char_data <= vdata ;
				pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
			when 3 =>		-- character count has done it's job and can be updated
				c_count <= c_count+1 ;
				pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
			when 4 =>		-- put the character data address ontot he bus
				Vaddr(11 downto 10) <= "11" ;
				Vaddr(9 downto 3) <= char_data(6 downto 0);
				Vaddr(2 downto 0) <= std_logic_vector(to_unsigned(scanLine,3));	-- will reloaad rows 0 and 1 at the bottom of the character but this is masked by the border
				pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
			when 5 =>
				pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
			when 6 =>
				pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
			when 7 =>		-- read the pixel data from the chatater map
				if char_data(7) ='0' then
					pixel_data <= vdata ;
				else
					pixel_data <= not(vdata) ;
				end if ;
		end case ;	
	end if;
end process ;

-- 4 control registers implemented
-- register 0 text colour
-- register 1 screen bacground colour
-- register 2 border/spare line colour
-- register 3 scroll/interrupr register
-- 				bit 7: 0 set vertical offset - 1 set horizontal offset
--					bit 6: 0 vbi disabled - 1 VBI enabled
--					bits 0-5: offset value for scrolling

configuartion:
process(reset,regUpdate)
begin
	if reset='0' then
		ink_data <= "11111111" ;
		paper_data <= "00000000" ;
		border_data <= "00000000" ;	
		vScrollReg <= "000000" ;
		hScrollReg <= "000000" ;
		VBIenable <='0' ;
	elsif rising_edge(regUpdate) then
		case Raddr is			
		when "00" =>
			ink_data 	<= Rdata ;
		when "01" =>
			paper_data 	<= Rdata ;
		when "10" =>
			border_data <= Rdata ;
		when "11" =>
			VBIenable <= Rdata(6) ;
			if Rdata(7)= '0' Then
				IF Rdata(5 downto 4) /= "11" Then
					vScrollReg <= Rdata(5 downto 0) ;
				end if;
			else
				hScrollReg <= Rdata(5 downto 0) ;
			end if;
		end case ;
	end if ;
end process;

-- shift register runs on the opposite edge of the clock to the memory
pixelOutput:
process(clk)
begin
	if falling_edge(clk) then
		if (border_area='1') then	
			if pixel_data(7)='1' then
				red_x 	<= ink_data(0) ;
				red_i 	<= ink_data(1) ;
				red 		<= ink_data(2) ;
				green_i 	<= ink_data(3) ;
				green 	<= ink_data(4) ;
				blue_x 	<= ink_data(5) ;
				blue_i 	<= ink_data(6) ;
				blue 		<= ink_data(7) ;		
			else
				red_x 	<= paper_data(0) ;
				red_i 	<= paper_data(1) ;
				red 		<= paper_data(2) ;
				green_i 	<= paper_data(3) ;
				green 	<= paper_data(4) ;
				blue_x 	<= paper_data(5) ;
				blue_i 	<= paper_data(6) ;
				blue 		<= paper_data(7) ;
			end if;
		else	
			if blank_area ='0' then
				red_x 	<= '0' ;
				red_i 	<= '0' ;
				red 		<= '0' ;
				green_i 	<= '0' ;
				green 	<= '0' ;
				blue_x 	<= '0' ;
				blue_i 	<= '0' ;
				blue 		<= '0' ;
			else
				red_x 	<= border_data(0) ;
				red_i 	<= border_data(1) ;
				red 		<= border_data(2) ;
				green_i 	<= border_data(3) ;
				green 	<= border_data(4) ;
				blue_x	<= border_data(5) ;
				blue_i	<= border_data(6) ;
				blue	 	<= border_data(7) ;
			end if ;
		end if;
	end if;
end process;

-- Horizontal Timing

-- Master horizontal counter, in character positions.
-- H_LineEnd is set for the last character position in the row, when this falls
-- again at the start of the next row, that triggers the vertical display actions 

HorizontalCount:
process(Clk, reset)
begin
    if (reset = '0') then
        H_Count <= 0;
    elsif rising_edge(Clk) then
        if H_Count = (H_Total - 1) then
            H_Count <= 0;
        else
             H_Count <= H_Count + 1;
        end if;
    end if;
end process;

-- The display is "active" up to the start of the front porch, after that is
-- either sync or porch until the other border starts, for the the output multiplexer
-- all non displayed area is the same, so only a generic "blank" is needed.
-- active low 

HorizontalDisplay:
process(Clk)
begin
    if rising_edge(Clk) then
        if (H_Count >= H_Porch) and (H_count < L_border) then
            H_Blank <= '0';
        else
            H_Blank <= '1';
        end if;
    end if;
end process;

-- the border area is 8 characters on the right, 8 on the left to centre 
-- the 64 active columns on the screen overlap with blank area dealt with elsewhere
-- active low
HorizontalBorder:
process(clk)
begin
    if rising_edge(clk) then
        if (H_Count >= R_border) then
            H_border <= '0';
        else
            H_border <= '1';
        end if;
    end if;
end process;

-- Horizontal sync is active in the blank part of the display
-- It has it's own output, so won't interfere with the blanking signal
-- active low
HorizontalSync:
process(clk)
begin
    if rising_edge(clk) then
        if ((H_Count >= H_Sync) and (H_Count< H_SyncEnd)) then
            HSYNC <= '0';
        else
            HSYNC <= '1';
        end if;
    end if;
end process;

--
-- Vertical Timing
-- similar to horizontal timing in that there are 4 types of display
-- sync, blank, border and active

-- Master vertical counter, counting in rows triggered by the line end pulse

VerticalCount:
process (clk, reset)
begin
	if (reset = '0') then
		V_Count <= 0;
	elsif rising_edge(clk) then
		if (h_count = h_sync) then
        if (V_Count = (V_Total-1)) then
            V_Count <= 0 ;
				rowCount <= to_integer(unsigned(vScrollReg))-1 ;
				scanline <= 9 ;
        else
            V_Count <= V_Count + 1;
				scanline <= scanline+1 ;
				if scanline > 8 then
					scanline <=0 ;
					rowCount <= rowCount +1 ;
					if rowCount > 46 then
						rowcount<= 0 ;
					end if ;
				end if ;
			end if;
		end if;
	end if;
end process;

-- Similar to the horizontal counter, the display is active from row 0, up to
-- the start of the porch area, where the blanking signal is set until the
-- border re-starts
-- active low

VerticalDisplay:
process(clk)
begin
    if rising_edge(clk) then
        if (V_Count >= V_porch) then
            V_Blank <= '0';
        else
            V_Blank <= '1';
        end if;
    end if;
end process;



-- Vertical sync is active for 2 full rows on the middle of the blanking period
-- active low

VerticalSync:
process(clk)
begin
    if rising_edge(clk) then
        if (V_Count>= V_Sync) and (V_Count < V_SyncEnd) then
            VSYNC <= '0';
      else
            VSYNC <= '1';
        end if;
    end if;
end process;

end behavioral;
